import java.awt.*;
import java.io.*;
import java.text.DecimalFormat;

public class procs
// class of procedures 
{
    private plots p = new plots();
    private outfile o;
    private map m = new map();
    private geom geom = new geom();
    private time t = new time();
    private sockport s;
    private double ylim = 415.0;
    private static DecimalFormat outS3;


    public procs(sockport sp, outfile out) {
        o = out;
        s = sp;
	outS3 = new DecimalFormat("0.###E0");
	outS3.setMinimumFractionDigits(3);
    }

    // take a spectrum; this gets called every time the program goes
    // through the main loop in srt.java
    // the first time called after startup, g.get_stow() should return 1
    // and azcmd and elcmd will be set to stow position
    void spectra(global g, disp d, Graphics gg)
    {
	int dm, i, k, j, n, nk, nk2, r, rn, sc;
	int nrows, pmax, ccol, crow, scol, srow, cshift, top, bot; //JDL
	double fshift, range; //JDL
        double freqf, secc, avp, a, pwr, pp, lst, slope;
	String astr = new String("*");
	dm = g.get_digital(); // digital mode?
	sc = g.get_scan(); // npoint scan if >= 1, bsw if < 0
//        if (sc > 0) r = g.get_rpt();
//        else r = 1;
	r = g.get_rpt();
	// first see if npoint contour map should be drawn
	if (g.get_map() == 2) { // draw the map when g.get_map() is 2
            m.map(g, d, gg, o);
	    // map() draws the coutour map and finds the peak position
	    // and width; it sets azoff and eloff to the peak position
	    // relative to the source (center of the npoint scan).
            g.set_map(1);
            g.set_azoff(0.0);
            g.set_eloff(0.0);
        }
	// if we're tracking, see if in a npoint or beamsw scan
        if (g.get_track() != 0) {
	    // g.set_scan(1) to start a npoint scan
	    // if starting a scan, check range JDL
	    if (sc == 1) {
		range = 0.25 * g.get_beamw() * ((double)g.get_grid() - 1.0);
		if (g.get_npaz()) {
		    if ((g.get_aznow() + g.get_pazoff() - range) < g.get_azlim1() ||
			(g.get_aznow() + g.get_pazoff() + range) > g.get_azlim2() ) {
			d.ppclear(gg, 8.0, ylim + 60.0, 350.0);
			d.dtext(8.0, ylim + 60.0, gg, Color.red,
			    "azimuth out of range");
			g.set_scan(0);
			return;
		    }
		}
		if (g.get_npel()) {
		    if ((g.get_elnow() + g.get_peloff() - range) < g.get_ellim1() ||
			(g.get_elnow() + g.get_peloff() + range) >= g.get_ellim2() ) {
			d.ppclear(gg, 8.0, ylim + 60.0, 350.0);
			d.dtext(8.0, ylim + 60.0, gg, Color.red,
			    "elevation out of range");
			g.set_scan(0);
			return;
		    }
		}
		d.set_bc(Color.green, 6);
	    }
	    // in a grid scan npaz & npel are both true and smax is grid^2
	    // pmax marks the end of a npoint scan
	    if (g.get_npaz() && g.get_npel()) {
		pmax = 1 + g.get_smax();
	    } else {
		pmax = 1 + g.get_grid();
	    }

	    if (sc >= pmax) { // has npoint scan finished ?
		if (g.get_fstatus() == 1) {
		    o.stroutfile(g, "* npoint finished");
		}
		g.set_scan(0); // stop the scans
		sc = 0;
                g.set_azoff(0);
                g.set_eloff(0);
		d.set_bc(Color.black, 6);
                m.map(g, d, gg, o);
		// map() leaves azoff and eloff (see note above) set
		// to the peak power position of the npoint scan;
		// topk can be changed in parameter window - JDL
		if (!g.get_topk()) {  // if topk is false, go back to
		    g.set_azoff(0.0); // the catalog target
		    g.set_eloff(0.0);
		}
		g.set_map(1); // so click can redraw it
            }
	    // do the next point in an npoint scan
	    // this involves calculating azoff, eloff
	    // JDL made changes to enable line or grid scans
	    if (sc != 0) {
		nrows = g.get_grid();
		if (!g.get_npel() && g.get_npaz()) { // azimuth
		    scol = sc;
		    srow = (nrows + 1)/2;
		    if (scol == 1) {
			d.ppclear(gg, 8.0, ylim + 60.0, 350.0);
//                        d.ppclear(gg, 8.0, ylim + 74.0, 200.0);
			if (g.get_fstatus() == 1) {
			    o.stroutfile(g, "* npoint azimuth scan ("
				+ g.get_sounam(g.get_sourn())
				+ ") of " + nrows + " points");
			    o.stroutfile(g,
				"* Roll-off = " + g.get_rlo() + " channels.");
			    if (r > 1) o.stroutfile(g,
				"* Each line is the average of " + r + " scans.");
			}
		    }
		    d.stext(8.0, ylim + 60.0, gg, Color.blue,
			"npoint azimuth, point "+scol+" of "+nrows+"; click npoint to abort");
		}
		else if (g.get_npel() && !g.get_npaz()) { // elevation
		    srow = sc;
		    scol = (nrows + 1)/2;
		    if (srow == 1) {
			d.ppclear(gg, 8.0, ylim + 60.0, 350.0);
//                        d.ppclear(gg, 8.0, ylim + 74.0, 200.0);
			if (g.get_fstatus() == 1) {
			    o.stroutfile(g, "* npoint elevation scan ("
				+ g.get_sounam(g.get_sourn())
				+ ") of " + nrows + " points");
			    o.stroutfile(g,
				"* Roll-off = " + g.get_rlo() + " channels.");
			    if (r > 1) o.stroutfile(g,
				"* Each line is the average of " + r + " scans.");
			}
		    }
		    d.stext(8.0, ylim + 60.0, gg, Color.blue,
			"npoint elevation, point "+srow+" of "+nrows+"; click npoint to abort");
		}
		else { // grid
		    srow = 1 + ((sc - 1) / nrows);
		    scol = 1 + ((sc - 1) % nrows);
		    if (scol == 1) {
			d.ppclear(gg, 8.0, ylim + 60.0, 350.0);
//                        d.ppclear(gg, 8.0, ylim + 74.0, 200.0);
			d.stext(8.0, ylim + 60.0, gg, Color.blue,
			    "npoint grid, row "+srow+" of "+nrows+"; click npoint to abort");
			if (g.get_fstatus() == 1) {
			    o.stroutfile(g, "* npoint grid scan ("
				+ g.get_sounam(g.get_sourn())
				+ ") row " + srow + " of " + nrows);
			    o.stroutfile(g,
				"* Roll-off = " + g.get_rlo() + " channels.");
			    if (r > 1) o.stroutfile(g,
				"* Each line is the average of " + r + " scans.");
			}
		    }
		}
                g.set_azoff(g.get_beamw() * 0.5 *
		    (double) (scol - (nrows + 1)/2) /
		    Math.cos(g.get_elcmd() * Math.PI / 180.0));
		g.set_eloff(g.get_beamw() * 0.5 * (double) (srow - (nrows + 1)/2));
		g.set_pwr(0.0, sc);
//                d.dtext(8.0, ylim + 74.0, gg, Color.blue,
//                    "azoff "+(scol - (nrows + 1)/2)
//                    +", eloff "+(srow - (nrows + 1)/2));
		g.set_scan(sc + 1);
		sc++;
	    } // and of moving to next point in npoint scan
	    if (g.get_bsw() != 0) { // go to next point in beamsw scan
                g.set_sig(-g.get_sig());
                if (g.get_sig() == -1) {
                    g.set_azoff(g.get_beamw() * g.get_bsw() / Math.cos(g.get_elcmd() * Math.PI / 180.0));
                    g.set_bsw(-g.get_bsw());
                } else
                    g.set_azoff(0.0);
	    } else { // not doing bsw scan
                g.set_sig(1);
	    }
        }
	// update sky plot, find what antenna AZ and EL should be set to
	p.plotbox(g, d, gg); // draw axes, labels for sky plotting area
	geom.setsounam(g, d, gg, t, true); // plot the catalog sources
	// set corrections for axis tilts
        g.set_elcor(geom.antiltel(g.get_azcmd(), g));
        g.set_azcor(geom.antiltaz(g.get_azcmd(), g.get_elcmd(), g));

	// not doing a calibration, so repoint antenna if required
	if (g.get_calon() == 0 && g.get_docal() == 0)
	{   // So far, spectra() has not looked at g.get_stow(),
	    // but azcmd and elcmd will be set to stow position on
	    // first call; s.azel moves the antenna to (azcmd, elcmd).
	    // If get_stow() is 1, s.azel() will drive the antenna
	    // to the stow position with counts of 5000.

	    // geom.setsounam(), called just above, sets azcmd to
	    // include azoff and pazoff, and does the analog for
	    // elcmd; thus s.azel(), below, will send antenna to
	    // the intended spot.

	    s.azel(g.get_azcmd() + g.get_azcor(), g.get_elcmd()
		+ g.get_elcor(), g, d, gg, o);

	    // scan was incremented above, so we use g.get_scan() - 1
	    // as the index below.
	    if (sc > 0) {
		// these are used in drawing the map
                a = g.get_aznow() - g.get_azcmd() + g.get_azoff();
		g.set_azscanoff(a * Math.cos(g.get_elcmd() * Math.PI / 180.0),
		    sc - 1);
		g.set_elscanoff(g.get_elnow() - g.get_elcmd()
		    + g.get_eloff(), sc - 1);
		ccol = g.get_grid() / 2;
		if (g.get_npaz() && g.get_npel())
		    crow = ccol;
		else
		    crow = 0;
		g.set_azhboff( (sc - 2) % g.get_grid() - ccol,
		    sc - 1);
		g.set_elhboff( (sc - 2) / g.get_grid() - crow,
		    sc - 1);
            }
        }
        if (g.get_track() == 0 && g.get_stow() != -1) {
	    return; // not at stow, so nothing more to do at this time
        }
        if (g.get_clr() != 0) {
            for (i = 0; i < g.get_nfreq(); i++) {
                g.set_avspec(0.0, i);
                g.set_avspecc(0.0, i);
            }
            g.set_av(0.0);
            g.set_avc(0.0);
            g.set_bswav(0.0);
            g.set_bswsq(0.0);
            g.set_bswlast(0.0);
            g.set_bswcycles(0);
            g.set_paver(0.0);
            g.set_pnum(1e-6);
            g.set_prms(0.0);
            if (g.get_clr() == -1)
                d.pclear(gg, 0.0, 0.0, 640.0, 440.0);
            g.set_clr(0);
        }
        if (g.get_track() == 1 && g.get_stow() == 0) {
            d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: tracking  ");
            d.set_bc(Color.green, 4);
        }
        if (g.get_track() == -1) {
            d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: stopped   ");
            d.set_bc(Color.red, 4);
        }
        if (g.get_track() == 0)
            d.set_bc(Color.black, 4);
	if (sc != 0) {
            d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: scanning");
            d.set_bc(Color.green, 6);
        } else
            d.set_bc(Color.black, 6);
        if (g.get_bsw() != 0) {
            d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: on/offs   ");
            d.set_bc(Color.green, 7);
	} else {
            d.set_bc(Color.black, 7);
	}
	// display number of gnd and radio communication errors
        if (g.get_comerr() > 0 || g.get_comerad() > 0)
            d.dtext(440.0, ylim + 40.0, gg, Color.blue, "ERRORS: " +
                    "gnd " + g.get_comerr() + " radio " + g.get_comerad());
        avp = 0.0;

	if (dm == 0 || dm == 5) {// modes 0 and 5, won't be called by students.
	    for (i = 0; i < g.get_nfreq(); i++) {
		if (g.get_mclick() == 1 || g.get_click() != 0) {
		    g.set_mclick(0);
		    return;
		}
		freqf = g.get_fcenter() + (double) (i - g.get_nfreq() / 2) * g.get_freqsep() + 0.8;
		g.set_spec(s.radiodg(freqf, g, d, gg, o), i); // get power from radio
		if (i == 0) g.set_freq0(g.get_freqa());
	    }
	    for (i = 0; i < g.get_nfreq(); i++) {
		pwr = g.get_spec(i);
		if (pwr > 0.0) {
		    avp += pwr;
		    if (g.get_sig() == 1)
			g.set_avspec(g.get_avspec(i) + pwr, i);
		    else
			g.set_avspecc(g.get_avspecc(i) + pwr, i);
		    if (sc != 0)
			g.set_pwr(g.get_pwr(sc - 1) + pwr, sc - 1);
		}
	    }
	}   // end of analog receiver (or digital modes 0 or 5) part
	else {
	// start digital receiver (mode != 5) part
	    bot = g.get_rlo();
	    top = 65 - bot;
	    cshift = 66 - (2 * bot);

	    fshift = Math.ceil(g.get_freqsep() * (double)(1000*cshift));
	    fshift /= 1000.0;  // now in MHz
	    //System.out.println("Frequency Shift: " + fshift);
	    if (dm == 4) {
		nk = 3; // nk is # scans to patch together
	    } else {
		nk = 1;
	    }

// this is so each spectrum scan can be the sum of r scans
	    for (rn = 1; rn <= r; rn++) {
		if (sc > 0) { // true if npoint scan
		    if (rn == 1) {
			nrows = g.get_grid();
			if (!g.get_npel() && g.get_npaz()) { // azimuth
			    scol = sc;
			    srow = (nrows + 1)/2;
			} else if (g.get_npel() && !g.get_npaz()) { // elevation
			    srow = sc;
			    scol = (nrows + 1)/2;
			} else { // grid
			    srow = 1 + ((sc - 1) / nrows);
			    scol = 1 + ((sc - 1) % nrows);
			}
			astr = "azoff "+(scol - (nrows + 1)/2)
			    +", eloff "+(srow - (nrows + 1)/2);
		    }
		    d.ppclear(gg, 8.0, ylim + 74.0, 350.0);
		    if (r ==  1) d.stext(8.0, ylim + 74.0, gg, Color.blue,
			astr);
		    if (r >  1) d.stext(8.0, ylim + 74.0, gg, Color.blue,
			astr+", scan "+rn+" of "+r);
		}
		if (rn % 3 ==  0) { // repoint antenna if needed
		    geom.setsounam(g, d, gg, t, false);
		    g.set_elcor(geom.antiltel(g.get_azcmd(), g));
		    g.set_azcor(geom.antiltaz(g.get_azcmd(), g.get_elcmd(), g));
		    s.azel(g.get_azcmd() + g.get_azcor(), g.get_elcmd()
			+ g.get_elcor(), g, d, gg, o);
		}
		// now take the spectrum
		for (k = 0; k < nk; k++) {
		    if (g.get_mclick() == 1 || g.get_click() != 0) {
			g.set_mclick(0);
			return;
		    }
		    nk2 = nk / 2;
		    freqf = g.get_fcenter() + (k - nk2) * fshift + 0.8;
		    //System.out.println("Frequency: " + freqf);
		    // 0.8 MHz I.F.
		    // center of middle scan is the center freq; the other
		    // two are centered cshift bins above and below that.

		    s.radiodg(freqf, g, d, gg, o); // get power from radio

		    if (k == nk2) // correct to actual setting
			g.set_fcenter(g.get_freqa());

		    // freq0 is the frequency of spec[0]
		    if (k == 0) {
			if (dm <= 3)
			    g.set_freq0(g.get_fcenter() - 32.0 * g.get_freqsep());
			else
			    g.set_freq0(g.get_fcenter() - 72.0 * g.get_freqsep());
		    }
		    for (i = 0; i < 64; i++) {
			pwr = g.get_specd(i);
			if (rn == 1) {
			    g.set_spec(pwr, i + k * 64);
			} else {
			    g.set_spec(pwr + g.get_spec(i + k * 64), i + k * 64);
			    if (rn == r)
				g.set_spec(g.get_spec(i + k * 64) / r, i + k * 64);
			}
		    }
// The 3 scans are now loaded into spec[0] thru spec[192];
// spec[96] is the center freq, spec[32] is cshift * freqsep
// lower, spec[160] is cshift * freqsep higher.
// Thus spec[32+cshift/2] and spec[95-cshift/2] are freqsep apart,
// as are spec[96+cshift/2] and spec[159-cshift/2].
// With respect to center freq (spec[96]), spec[63-rlo] is
// cshift/2 bins down, spec[97-cshift/2] is cshift/2 - 1 bins down,
// spec[96+cshift/2] is cshift/2 bins up, and spec [161-cshift/2]
// is cshift/2 + 1 bins up.
		} // end of k for loop for mode 4
	    } // end of rn for loop to repeat scans at each  npoint
	    if (dm == 4) {
		if (g.get_match() == 1) {
		    pwr = 0.5*(g.get_spec(64+bot) + g.get_spec(63+bot)
			  - g.get_spec(top) - g.get_spec(top+1));
		    for (i = 0; i < 64; i++) {
			g.set_spec(g.get_spec(i) + pwr, i);
		    }
		    pwr = 0.5*(g.get_spec(128+bot) + g.get_spec(127+bot)
			  - g.get_spec(top+64) - g.get_spec(top+65));
		    for (i = 128; i < 192; i++) {
			g.set_spec(g.get_spec(i) - pwr, i);
		    }
		} else if (g.get_match() == 2) {
		    pwr = (g.get_spec(64+bot) + g.get_spec(63+bot))
			/ (g.get_spec(top) + g.get_spec(top+1));
		    for (i = 0; i < 64; i++) {
			g.set_spec(g.get_spec(i) * pwr, i);
		    }
		    pwr = (g.get_spec(top+64) + g.get_spec(top+65))
			/ (g.get_spec(128+bot) + g.get_spec(127+bot));
		    for (i = 128; i < 192; i++) {
			g.set_spec(g.get_spec(i) * pwr, i);
		    }
		} else {
		    pwr = 0.0;
		    for (k = 0; k < 3; k++) {
			i = 64 - bot + k;
			j = 64 + bot + k;
			pwr += g.get_spec(i) - g.get_spec(j);
			i = 128 - bot + k;
			j = 128 + bot + k;
			pwr += g.get_spec(i) - g.get_spec(j);
		    }
		    slope = pwr / (double)(6*cshift);
		    for (k = 0; k < 3; k++) {
			for (i = 0; i < 64; i++) {
			    j = i + k * 64;
			    pwr = g.get_spec(j) - (i - 32) * slope;
			    g.set_spec(pwr, j);
			}
		    }
		}
	    } // end of if digital == 4 section
	    for (k = 0; k < nk; k++) {
		for (i = 0; i < 64; i++) {
		    j = 1;
		    n = cshift * k + i;
		    if (k == 0 && dm == 4) {
			if (i > top)
			    j = 0;
		    }
		    if (k == 1) {
			if (i < bot || i > top)
			    j = 0;
		    }
		    if (k == 2) {
			if (i < bot)
			    j = 0;
		    }
// If j==1, the channel is used to make the patched wide scan.
// It consists of spec[0] to spec[65-rlo]
// from the low scan, channels spec[64+rlo] to spec[64+65-rlo]
// from the mid scan, channels spec[128+rlo] to spec[191]
// from the hi scan.
// For mode 1, k is only 0; there is no change in spec[]
// and avp is computed from spec[rlo] thru spec[63-rlo].
// For mode 4, elements in the patched wide scan with index from rlo
// to 195-5*rlo will be used to find avp.
		    if (j == 1) {
			pwr = g.get_spec(i + k * 64);
			if (dm == 4) // otherwise, n == i
			    g.set_spec(pwr, n);
			if (pwr > 0.0) {
			    if (n >= bot && n < g.get_nfreq() - bot) {
				avp += pwr;
				if (sc != 0)
				    g.set_pwr(g.get_pwr(sc - 1) + pwr, sc - 1);
			    }
			    if (g.get_sig() == 1)
				g.set_avspec(g.get_avspec(n) + pwr, n);
			    else
				g.set_avspecc(g.get_avspecc(n) + pwr, n);
			}
		    }
		}
	    }
	} // end of digital receiver part

	// compute average power for the scan
	if (dm > 0) // JDL
	    avp = avp / ((double) g.get_nfreq() - 24); // rolloff?
        else
            avp = avp / (double) g.get_nfreq();
        if (g.get_bswlast() > 0.0) {
            if (g.get_sig() == 1)
                pp = avp - g.get_bswlast();
            else
                pp = g.get_bswlast() - avp;
            g.set_bswav(g.get_bswav() + pp);
            g.set_bswsq(g.get_bswsq() + pp * pp);
            g.set_bswcycles(g.get_bswcycles() + 1);
        }
        g.set_bswlast(avp);
        if (g.get_sig() == 1)
            g.set_av(g.get_av() + 1);
        else
            g.set_avc(g.get_avc() + 1);
        if (g.get_clr() != 0) {
            for (i = 0; i < g.get_nfreq(); i++) {
                g.set_avspec(0.0, i);
                g.set_avspecc(0.0, i);
            }
            g.set_av(0.0);
            g.set_avc(0.0);
            g.set_paver(0.0);
            g.set_prms(0.0);
            g.set_pnum(1e-6);
            g.set_clr(0);
        }
	p.plotspec(0, g, d, gg, o); // does the black spectral plot
	p.plotspec(1, g, d, gg, o); // does the red spectral plot
	// current time; sent only to plotp(), which ignores it.
	secc = (double) t.getTsec(g, d, gg);
	// update the time output box
        lst = (t.getGst() - g.get_lon()) * 12.0 / Math.PI;
        if (lst < 0.0)
            lst += 24.0;
        if (lst > 24.0)
            lst -= 24.0;
        d.dtext(670.0, 167.0, gg, Color.black, "LST " + d.dc(lst, 5, 2) + " hrs");
	d.ppclear(gg, 710.0, 136.0, 250.0);
        d.dtext(710.0, 136.0, gg, Color.black, "UTdate " + t.getMonthday() + " ");
        if (g.get_tsys() > 0.0)
            d.dtext(665.0, 328.0, gg, Color.black, "tsys: " + d.dc(g.get_tsys(), 5, 0) + " K             ");
        else
            d.dtext(665.0, 328.0, gg, Color.black, "tsys: not yet calibrated ");
	if (sc == 0 && g.get_stow() == 0 && g.get_track() == 0)
            d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status:                ");
        if (g.get_track() == 0)
            d.set_bc(Color.black, 4);
	p.plotp(secc, g, d, gg); // updates power vs time plot
        if (g.get_fstatus() == 1 && g.get_recmode() != 4)
	    // write the scan results to the record file
	    o.writeoutfile(g, d, gg);
    }

    // vane calibration procedure
    void calibrat(global g, disp d, Graphics gg, outfile o)
    {
        int i, istart, istop;
        double pwr0, pwr1, trecvr;
        g.set_atten(0);         /* turn-off atten for calibration */
        d.set_bc(Color.green, 14);
        pwr0 = pwr1 = 0.0;
        if (g.get_mancal() == 0)
            s.cal(1, g, d, gg, o);
        else {
            d.dtext(400.0, ylim + 60.0, gg, Color.red, "place load on feed");
            g.set_stopproc(1);
            while (g.get_stopproc() == 1) {
                if (g.get_click() == 1) {
                    g.set_stopproc(0);
                    g.set_click(0);
                }
            }
            d.dtext(400.0, ylim + 60.0, gg, Color.white, "place load on feed");
        }
        g.set_calon(1);
        spectra(g, d, gg);
        istart = 0;
        istop = g.get_nfreq();
        if (g.get_digital() > 0) {
            istart = 10;
            istop = g.get_nfreq() - 10;
        }
        for (i = istart; i < istop; i++)
            pwr1 += g.get_spec(i);
        pwr1 = pwr1 / (double) (istop - istart);
        if (g.get_mancal() == 0)
            s.cal(0, g, d, gg, o);
        else {
            d.dtext(400.0, ylim + 60.0, gg, Color.red, "remove load");
            g.set_stopproc(1);
            while (g.get_stopproc() == 1) {
                if (g.get_click() == 1) {
                    g.set_stopproc(0);
                    g.set_click(0);
                }
            }
            d.dtext(400.0, ylim + 60.0, gg, Color.white, "remove load");
        }
        g.set_calon(0);
        spectra(g, d, gg);
        for (i = istart; i < istop; i++)
            pwr0 += g.get_spec(i);
        pwr0 = pwr0 / (double) (istop - istart);
        if (pwr1 > pwr0 && pwr0 > 0.0) {
            trecvr = (g.get_tload() - (pwr1 / pwr0) * g.get_tspill()) / (pwr1 / pwr0 - 1.0);
            g.set_tsys(trecvr + g.get_tspill());
            g.set_calcons((trecvr + g.get_tspill()) * g.get_calcons() / pwr0);
            d.dtext(665.0, 328.0, gg, Color.black, "tsys: " + d.dc(g.get_tsys(), 5, 0) + " K             ");
//            d.dtext(665.0, 344.0, gg, Color.black, "calcons: " + d.dc(g.get_calcons(), 5, 2));
	    d.dtext(665.0, 344.0, gg, Color.black, "calcons: " + outS3.format(g.get_calcons()));
            d.dtext(665.0, 360.0, gg, Color.black, "trec: " + d.dc(trecvr, 5, 0) + " K");
            if (g.get_fstatus() == 1)
                o.stroutfile(g,
//                             "* tsys " + d.dc(g.get_tsys(), 5, 0) + " calcons" + d.dc(g.get_calcons(), 5, 3) +
			     "* tsys " + d.dc(g.get_tsys(), 5, 0) + " calcons" + outS3.format(g.get_calcons()) +
                             " trecvr" + d.dc(trecvr, 5, 0) + " tload" + d.dc(g.get_tload(), 5, 0) +
                             " tspill" + d.dc(g.get_tspill(), 5, 0));
            g.set_docal(0);
            d.set_bc(Color.black, 14);
        } else {
	    if (g.get_fstatus() == 1) // JDL
		o.stroutfile(g, "* tsys: error");
            d.dtext(670.0, 296.0, gg, Color.black, "tsys: error");
            g.set_docal(0);
            d.set_bc(Color.black, 14);
        }
    }

    // noise diode calibration procedure
    void noisecal(global g, disp d, Graphics gg, outfile o)
    {
        int i, istart, istop;
        double pwr0, pwr1, trecvr;
        d.set_bc(Color.green, 13);
        pwr0 = pwr1 = 0.0;
        g.set_atten(0);
        s.cal(3, g, d, gg, o);
        g.set_calon(2);
        spectra(g, d, gg);
        istart = 0;
        istop = g.get_nfreq();
        if (g.get_digital() > 0) {
            istart = 10;
            istop = g.get_nfreq() - 10;
        }
        for (i = istart; i < istop; i++)
            pwr1 += g.get_spec(i);
        pwr1 = pwr1 / (double) (istop - istart);
        s.cal(2, g, d, gg, o);
        g.set_calon(0);
        g.set_atten(0);         /* turn-off atten for calibration */
        spectra(g, d, gg);
        for (i = istart; i < istop; i++)
            pwr0 += g.get_spec(i);
        pwr0 = pwr0 / (double) (istop - istart);
        if (pwr1 > pwr0 && pwr0 > 0.0 && g.get_noisecal() > 0.0) {
            trecvr = ((g.get_noisecal()) / (pwr1 / pwr0 - 1.0)) - g.get_tspill();
            g.set_tsys(trecvr + g.get_tspill());
            g.set_calcons((trecvr + g.get_tspill()) * g.get_calcons() / pwr0);
            d.dtext(665.0, 328.0, gg, Color.black, "tsys: " + d.dc(g.get_tsys(), 5, 0) + " K             ");
//            d.dtext(665.0, 344.0, gg, Color.black, "calcons: " + d.dc(g.get_calcons(), 5, 2));
	    d.dtext(665.0, 344.0, gg, Color.black, "calcons: " + outS3.format(g.get_calcons()));
            d.dtext(665.0, 360.0, gg, Color.black, "trec: " + d.dc(trecvr, 5, 0) + " K");
	    if (g.get_fstatus() == 1) {
                o.stroutfile(g,
		    "* tsys " + d.dc(g.get_tsys(), 5, 0) + " calcons" + outS3.format(g.get_calcons()) +
		    " trecvr" + d.dc(trecvr, 5, 0) + " tload" + d.dc(g.get_noisecal(), 5, 0) +
		    " tspill" + d.dc(g.get_tspill(), 5, 0));
                o.stroutfile(g,
		    "* Roll-off = " + g.get_rlo() + " channels.");
	    }
            g.set_docal(0);
            d.set_bc(Color.black, 13);
        } else {
	    if (g.get_fstatus() == 1) // JDL
		o.stroutfile(g, "* tsys: error");
            d.dtext(670.0, 296.0, gg, Color.black, "tsys: error");
            g.set_docal(0);
            d.set_bc(Color.black, 13);
        }
    }

    void grayfilt(global g) {
        double cf[] = {
            1.000000, 1.006274, 1.022177, 1.040125, 1.051102, 1.048860, 1.033074, 1.009606,
            0.987706, 0.975767, 0.977749, 0.991560, 1.009823, 1.022974, 1.023796, 1.011319,
            0.991736, 0.975578, 0.972605, 0.986673, 1.012158, 1.032996, 1.025913, 0.968784,
            0.851774, 0.684969, 0.496453, 0.320612, 0.183547, 0.094424, 0.046729, 0.026470,
            0.021300
        };
        int i;
        for (i = 0; i <= 32; i++) {
            if (i < 32)
                g.set_graycorr(cf[i], i + 32);
            if (i < 33)
                g.set_graycorr(cf[i], 32 - i);
        }
    }
}
