import java.awt.*;
import java.util.*;
import java.text.*;
import java.lang.*;
import java.io.*;
import java.net.*;
import javax.swing.*;

public class sockport
// socketport class - talks to the motor and radio
// microcontrollers over TCP socket connection and serial port.
// Must talk to one controller at a time and get a response
// before talking again.
{
  private geom geom = new geom();
  private time t = new time();
  static BufferedReader inReader;
  static PrintWriter outWriter = null;
  static Socket spSocket = null;
  private double ylim = 415.0;
  private double ylim0 = 200.0;
  private int azatstow = 0;
  private int elatstow = 0;
//  private static BufferedWriter lout = null;  //JDL
  private static PrintStream lout = null;  //JDL

  public sockport(global g)
  {
      int chose = 10;    // Not a JOptionPane return option
      int essai = 0;
      Object[] choices = { "YES", "EXIT" };
      if (g.get_elog()) { //JDL set up error logging
	  if ((lout = g.get_elout()) == null) {
	      lout = g.open_elog(); // just returns pointer if already open
	  }
      } else {
	  lout = null;
	  System.setErr(g.get_eps());
      }
      if (g.get_azelsim() != 0 && g.get_radiosim() != 0)
	  return; // everything is simulated internally
      while (essai < 2) {
	  String status = "";
	  try {
	      spSocket = new Socket(g.get_host(), g.get_port());
	      spSocket.setTcpNoDelay(true);
	      spSocket.setSoLinger(true, 10);
	      outWriter = new PrintWriter(spSocket.getOutputStream(), true);
	      inReader = new BufferedReader(
		  new InputStreamReader(
		      spSocket.getInputStream()));
	      outWriter.println("RUBUSY?");
	      status = inReader.readLine();
//              System.out.println("Server replied: "+status);
	      // see srtdaemon.java for why we have this stuff
	      if (status.startsWith("NO")) {
		    // we are connected and to the right thread, so
		    // spSocket is good to use
//                  java.awt.Toolkit.getDefaultToolkit().beep();
		    essai = 4; // breaks us out of the loop
	      }
	      else if (status.startsWith("YES")) {
		   essai = 2;
		   chose = JOptionPane.showOptionDialog(null,
		       "Sorry, "+status.substring(4)+" is using the SRT.\n"
		       +"Turn on simulation?",
		       "SRT is Busy:",
		       JOptionPane.YES_NO_OPTION,
		       JOptionPane.PLAIN_MESSAGE,
		       null, choices, choices[0]);
	      }
	      else if (status.startsWith("403")) {
		  essai = 3;
		  chose = JOptionPane.showOptionDialog(null,
		      "Server says MIT only.\n"
		      +"Turn on simulation?",
		      "403 FORBIDDEN:",
		      JOptionPane.YES_NO_OPTION,
		      JOptionPane.PLAIN_MESSAGE,
		      null, choices, choices[0]);
	      }
	      if (essai == 0 || status.startsWith("MAYBE")) {
		  // go around again
		  outWriter.println("bye. ");
		  outWriter.flush();
		  outWriter.close();
		  outWriter = null;
		  inReader.close();
		  inReader = null;
		  spSocket.close();
		  spSocket = null;
		  try { Thread.sleep(10000); }
		  catch(InterruptedException e) { ; }
		  essai++;
	      }
	  } catch (UnknownHostException e) {
	      System.err.println("Don't know about host: " + g.get_host());
	      System.exit(1);
	  } catch (IOException e) {
	      System.err.println("Couldn't get I/O for the connection to " + g.get_host());
	      JOptionPane.showMessageDialog(null,
		  "Can't connect to "+g.get_host()
		  +".\nIs the server running?",
		  "SRT Server:",
		  JOptionPane.ERROR_MESSAGE);
	      System.exit(1);
	  }
	  if (chose == JOptionPane.YES_OPTION) { // this is 0
	      g.set_azelsim(-1);
	      g.set_radiosim(1);
	      System.err.println("Cannot connect, simulating.");
	  } else if (chose != 10) {
	      System.err.println("Exiting rather than simulate.");
	      System.exit(0);
	  }
      }   // bottom of while (essai < 2) loop
/*
      if (lout != null) {
	  try {
	      lout.println(t.getYdhms() + " host " + g.get_host()
		+ " port " + g.get_port());
	      lout.flush();
	  } catch(IOException e) {
	      System.err.println(e);
	  }
      }
*/
  }  // end of sockport() initialization method; if we get to here
     // we are either simulating or spSocket is useable

  // method commands antenna movement to az, el
  // when this class is instantiated by new sockport(), azatstow and elatstow
  // will be zero.
  void azel(double az, double el, global g, disp d, Graphics gg, outfile o)
  {
    int i,k, kk, n, mm, ax, axis, count, ccount, rcount, fcount,
	b2count, region1, region2, region3, flip;
    double azscale, elscale, azz, ell, ellcount, ellnow, lenzero,
	ra, dec, glat, glon, vel, sec, x, y;
    int j;
    String str;
    String rstr = new String("");
    String[] part = {"1", "2", "3"};
    mm = count = 0;
    lenzero = 0.0;
    if (g.get_elog()) { //JDL
	if ((lout = g.get_elout()) == null) {
	  lout = g.open_elog();
	}
    } else {
	lout = null;
    }

    d.dtext(670.0, 24.0, gg, Color.black, "cmd " + d.dc(az, 5, 1) + " " + d.dc(el, 5, 1) + " deg");
    if (g.get_azelsim() == 0) {
	str = "antenna drive status:";
	if (g.get_comerr() > 0)
	    str += " comerr= " + g.get_comerr();
    } else
	str = "antenna drive simulated";
    if (g.get_vsrt() == 0)
	d.dtext(16.0, 32.0, gg, Color.black, str);
    // derive a strategy for getting to az el
    az = az % 360;          /* Fold into reasonable range */
    if (g.get_south() == 0) { //not the case for CASSI
	az = az + 360.0;    /* put az in range 180 to 540 */
	if (az > 540.0)
	    az -= 360.0;
	if (az < 180.0)
	    az += 360.0;
    }
    region1 = region2 = region3 = 0;
    if (az >= g.get_azlim1() && az < g.get_azlim2() && el >= g.get_ellim1()
      && el <= g.get_ellim2()) // target is between az limits
	region1 = 1;
    if (az > g.get_azlim1() + 180.0 && el > (180.0 - g.get_ellim2()))
	region2 = 1; // target in 2nd 1/2 az range and outside el range
    if (az < g.get_azlim2() - 180.0 && el > (180.0 - g.get_ellim2()))
	region3 = 1; // target in 1st 1/2 az range and outside el range
    if (region1 == 0 && region2 == 0 && region3 == 0) {
	d.dtext(16.0, 48.0, gg, Color.red, "cmd out of limits");
	d.set_bc(Color.black, 4);
	if (g.get_fstatus() == 1 && g.get_track() != 0)
	    o.stroutfile(g, "* ERROR cmd out of limits");
	g.set_track(0);
	try { Thread.sleep(100); }
	catch(InterruptedException e) { System.err.println(e); }
	return;
    }
    flip = 0;
    if (az > g.get_azlim2() && g.get_pushrod() == 0) {
	az -= 180.0;
	el = 180.0 - el;
	flip = 1;
    }
    if (az < g.get_azlim1() && g.get_pushrod() == 0 && flip == 0) {
	az += 180.0;
	el = 180.0 - el;
	flip = 1;
    }
    azz = az - g.get_azlim1(); // degrees from lower limit
    ell = el - g.get_ellim1(); // degrees from lower limit
//    scale = 52.0 * 27.0 / 120.0;
// Cassi mount azimuth motor is about 4.77 counts/degree
    azscale = g.get_azcounts_per_deg(); // converts azz to axis==0 counts
    elscale = g.get_elcounts_per_deg(); // converts ell to axis==1 counts
// mm==1 to increase az (CW) mm==0 to decrease az (CCW)
// mm==3 to increase el      mm==2 to decrease el
    g.set_slew(0);
    // find how many counts el motor must make (trigonometry)
    if (g.get_pushrod() == 0)
	ellcount = ell * elscale;
    else {
	lenzero = g.get_rod1() * g.get_rod1() + g.get_rod2() * g.get_rod2()
	  - 2.0 * g.get_rod1() * g.get_rod2()
	  * Math.cos((g.get_rod4() - g.get_ellim1()) * Math.PI / 180.0)
	  - g.get_rod3() * g.get_rod3();
	if (lenzero >= 0.0)
	  lenzero = Math.sqrt(lenzero);
	else
	  lenzero = 0;
	ellcount = g.get_rod1() * g.get_rod1() + g.get_rod2() * g.get_rod2()
	  - 2.0 * g.get_rod1() * g.get_rod2()
	  * Math.cos((g.get_rod4() - el) * Math.PI / 180.0)
	  - g.get_rod3() * g.get_rod3();
	if (ellcount >= 0.0)
	  ellcount = (-Math.sqrt(ellcount) + lenzero) * g.get_rod5();
	else
	  ellcount = 0;
// increase in rod length (mm == 2) drives towards horizon
//        System.out.println("ellcount "+ellcount);
    }
    if (ellcount > g.get_elcount() * 0.5)
	axis = 1; // move in elevation first
    else
	axis = 0; // move in azimuth first
    for (ax = 0; ax < 2; ax++) {
	// first find how many counts to move, keep count > 0
	// azcount > 0 is twice the current position in motor counts;
	// away from stow position -- similar for elcount
	if (axis == 0) {
	    if (azz * azscale > g.get_azcount() * 0.5 - 0.5) {
	      mm = 1; // move away from stow az
	      count = (int) (azz * azscale - g.get_azcount() * 0.5 + 0.5);
	    }
	    if (azz * azscale <= g.get_azcount() * 0.5 + 0.5) {
	      mm = 0; // move towards from stow az
	      count = (int) (g.get_azcount() * 0.5 - azz * azscale + 0.5);
	    }
	} else {
	    if (ellcount > g.get_elcount() * 0.5 - 0.5) {
	      mm = 3; // move away from stow el
	      count = (int) (ellcount - g.get_elcount() * 0.5 + 0.5);
	    }
	    if (ellcount <= g.get_elcount() * 0.5 + 0.5) {
	      mm = 2; // move towards from stow el
	      count = (int) (g.get_elcount() * 0.5 - ellcount + 0.5);
	    }
	}
	ccount = count;
	// if target is stow position - get_stow() is 1 - head towards
	// stow until the limit switches stop us
	// on first call of azel() after normal startup, azatstow, elatstow
	// will be zero, get_stow returns 1 and target will be stow position
	if (g.get_stow() == 1 && g.get_azcmd() == g.get_azstow()
	  && g.get_elcmd() == g.get_elstow()) //JDL
	{
	    if (g.get_fstatus() == 1)
		o.stroutfile(g, "* moving to stow");
	    count = 5000; // extra steps (b2count) are ignored
	    if (axis == 0) {
		mm = 0;
		if (azatstow == 1)
		    count = 0;
	    }
	    if (axis == 1) {
		mm = 2;
// complete azimuth motion to stow before completely dropping in elevation
		if (elatstow == 1 || (ccount <= 2.0 * g.get_countperstep()
		    && azatstow == 0))
		    count = 0;
	    }
	    flip = 0;
	}
	// JDL
	// this keeps count == 5000 when going to stow, no matter what stepping
	// size has been set; stepping is not implemented in an npoint scan
	if (count != 5000 && g.get_scan() == 0 && count > g.get_countperstep())
	    count = g.get_countperstep();

	// "ignore counts less than tolerance (default 1)
	if (count >= g.get_ptoler() && g.get_track() != -1) {
	  if (count > g.get_ptoler()) {
	      g.set_slew(1);
	      d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: slewing   ");
	      d.set_bc(Color.black, 4);
	      d.set_bc(Color.black, 3);
	  }
	  x = g.get_xlast(500);
	  y = g.get_ylast(500);
	  if (x != g.get_xlast(0) || y != g.get_ylast(0)) {
	      d.lpaint(gg, Color.white, (double) (x - 10), (double) y, (double) (x + 10), (double) y);
	      d.lpaint(gg, Color.white, (double) x, (double) (y - 10), (double) x, (double) (y + 10));
	  }
	  x = (int) (g.get_azcmd() * 640.0 / 360.0);
	  if (g.get_south() == 0)
	      x -= 320;
	  if (x < 0)
	      x += 640;
	  if (x > 640)
	      x -= 640;
	  y = (int) (ylim - g.get_elcmd() * (ylim - ylim0) / 90);
	  g.set_xlast(x, 500);
	  g.set_ylast(y, 500);
	  // changed from yellow to blue (easier to see) JDL
	  d.lpaint(gg, Color.blue, (double) (x - 10), (double) y, (double) (x + 10), (double) y);
	  d.lpaint(gg, Color.blue, (double) (x), (double) (y - 10), (double) (x), (double) (y + 10));
	  // no space at start, but a space needed after each number
	  str = "move " + mm + " " + count + "  ";
	  n = kk = 0;
	  if (g.get_azelsim() != 0) {
	      if (count < 5000)
		  rstr = new String("M " + count + " 0 " + mm + " \n");
	      else
		  rstr = new String("T "+count+" "+mm+" "+count+" \n");
	      n = rstr.length();
	      part = rstr.split("\\s", 0);
	      kk = part.length;
	  }
	  d.dtext(16.0, 64.0, gg, Color.black, "motor: " + str.substring(0, str.length() - 1) + "     ");
	  d.ppclear(gg, 16.0, 80.0, 180.0);
	  // For the Cassi Mount, there are about 1650 az motor steps
	  // between the limits, and 850 el steps. I measured about 5 motor
	  // steps per second, 500 s seems to be a reasonable timeout.
	  if   (g.get_azelsim() == 0) { // real
	      try {
		  // this is in units of 0.001s; make it long enough as the
		  // socket code won't return until the motor response is complete.
		  if (count > 1250)
		      spSocket.setSoTimeout(500000);
		  else if (count > 25)
		      spSocket.setSoTimeout(400*count);
		  else
		      spSocket.setSoTimeout(10000);
	      }
	      catch(SocketException e) {
		  System.err.println("move: can't set timeout "+e);
	      }
	      try { // start stepping
		  outWriter.println(str);
		  d.ppclear(gg, 16.0, 48.0, 180.0);
		  if (axis == 0)
		    d.dtext(16.0, 48.0, gg, Color.black, "waiting on azimuth   ");
		  else
		    d.dtext(16.0, 48.0, gg, Color.black, "waiting on elevation ");
		  try {
		      rstr = new String(inReader.readLine());
		      // this could be made more sophisticated
		      while (rstr.startsWith("BEEP")) {
			  g.set_beep(true);
			  rstr = new String(inReader.readLine());
		      }
		      j = (int) rstr.charAt(0);
		      if (j == 'M' || j == 'T')
			  outWriter.println("GotIt");
		  } catch(SocketTimeoutException e) { // extends InterruptedIOException
		      System.err.println("move "+mm+": read() timeout");
		  }
		  n = rstr.length();
		  if (n > 6) {
		      part = rstr.split("\\s", 0);
		      kk = part.length;
		  }
//                  t.getTsec(g, d, gg);
		  d.ppclear(gg, 16.0, 48.0, 180.0);
	      } catch(IOException e) {
		  System.err.println(e);
	      }
	  }
	  if (n <= 4 || kk < 3 || (rstr.charAt(0) != 'M' && rstr.charAt(0) != 'T')) {
	      d.dtext(16.0, 16.0, gg, Color.red, "comerr n="+n+" mtr="+mm+" cnt="+count);
	      d.dtext(16.0, 80.0, gg, Color.red, "reply: " + rstr);
	      g.set_comerr(g.get_comerr() + 1);
	      if (g.get_fstatus() == 1)
		  o.stroutfile(g, "* ERROR comerr");
	      if (lout != null) {
/*
		  try {
		      lout.write(t.getYdhms()+" comerr, got rstr = "+rstr);
		      lout.newLine();
		      lout.write("\tgoing to stow position");
		      lout.newLine();
		      lout.flush();
		  } catch(IOException e) {
		      System.err.println(e);
		  }
*/
		  lout.println(t.getYdhms()+" comerr, got rstr = "+rstr);
		  lout.println("\tgoing to stow position");
		  lout.flush();
	      }
	      if (g.get_mainten() == 0) {
		  if (g.get_fstatus() == 1)
		      o.stroutfile(g, "* commerr, going to stow, got rstr = "+rstr);
		  g.set_clr(1); //JDL
		  g.set_bsw(0); //JDL
		  g.set_scan(0); //JDL
		  g.set_sourn(0); //JDL
		  g.set_track(0); //JDL
		  g.set_azoff(0.0); //JDL
		  g.set_eloff(0.0); //JDL
		  g.set_azcmd(g.get_azstow());
		  g.set_elcmd(g.get_elstow());
		  g.set_stow(1); //JDL
	      }
	      return; // to procs.spectra() -- the only function that calls us
	  }

	  if (g.get_azelsim() != 0 && g.get_azelsim() < 10) {
	      d.ppclear(gg, 16.0, 48.0, 180.0);
	      try {
		  Thread.sleep(100);
	      } catch(InterruptedException e) {
		  System.err.println(e);
	      }
	  }
/* rstr is the information returned by the Stamp 1 motor controller.
 * With the older code (mtrnew4.bas) it will be:
 *      "E B8 W5"       if B8 was invalid (0-7 are OK)
 *      "M W2 W5 B8"    return from motor move command
 *      "T W2 B8 W5"    return from timeout
 *      "O W2 B8 W5"    return from CALOUT  (B8==4) command }
 *      "I W2 B8 W5"    return from CALIN   (B8==5) command } W2 should be 0
 *      "I W2 B8 W5"    return from NCALOFF (B8==6) command }
 *      "I W2 B8 W5"    return from NCALON  (B8==7) command }
 * where B8 specified the command, W5 was the steps requested,
 * and W2 was the steps taken (reed switch pulses).
 * B8==0 is M1CW (AZ), B8==1 is M1CCW (AZ),
 * B8==2 is M2CW (EL), B8==3 is M2CCW (EL),
 *
 * The newer code (mtrnew5.bas) returns:
 *      "E B8 W5"       if invalid B8
 *      "M W2 B2 B8"    return from motor move command
 *      "T W2 B8 W5"    return from timeout
 *      "C W2 B8 W5"    return from CALOUT  (B8==4) command }
 *      "C W2 B8 W5"    return from CALIN   (B8==5) command } W2 should be 0
 *      "C W2 B8 W5"    return from NCALOFF (B8==6) command }
 *      "C W2 B8 W5"    return from NCALON  (B8==7) command }
 * where the new variable B2 is the number of extra steps the
 * motor took; i.e., the reed switch returned W5 + B2 pulses.
 *
 * CALIN and NCALON both apply power to the noise calibrator.
 * CALIN is intended to run the vane motor in; it makes pin 3 of the
 * 9-pin CPC negative to move the vane in (this would forward bias
 * the avalanche noise diode) waits 20 sec and turns off the power.
 * NCALON is intended to turn on the noise diode; it makes pin 3 of
 * the 9-pin CPC positive (correct reverse bias to the diode),
 * leaves the voltage on, and returns.
 * CALOUT is intended to run the vane motor out; it makes pin 3 of the
 * the 9-pin CPC positive to move the vane out of the beam, waits 20 sec
 * and turns off the power.
 * NCALOFF is intended to turn off the noise diode; it turns off
 * the voltage, and returns.
 *
 */
	  d.ppclear(gg, 16.0, 80.0, 180.0);
	  d.dtext(16.0, 80.0, gg, Color.black, "reply: " + rstr);
//          part = rstr.split("\\s", 0); now split above
	  try {
	      rcount = Integer.parseInt(part[1]);
	  } catch(NumberFormatException fe) {
	      System.err.println(part[1] + ": " + fe);
	      rcount = 0;
	  }
	  if (g.get_stamp()) { //JDL, if true Stamp code is old, mtrnew4.bas
	      fcount = 0;
	  } else {             // Stamp code is new, mtrnew5.bas
	      try {
		  b2count = Integer.parseInt(part[2]);
	      } catch(NumberFormatException fe) {
		  System.err.println(part[2] + ": " + fe);
		  b2count = 0;
	      }
	      if (count < 5000) // add extra 1/2 count from motor coast
		  fcount = count * 2 + b2count;
	      else
		  fcount = 0;
	      // count is always 5000 when going to stow position

/* this clutters up the log file too much
	      if (b2count != 0 && lout != null) {
		  try {
		      lout.println(t.getYdhms()+" b2count: "+b2count);
//                      lout.write(t.getYdhms()+" b2count: "+b2count);
//                      lout.newLine();
		      lout.flush();
		  } catch(IOException e) {
		      System.err.println(e);
		  }
	      }
*/
	  }
	  // report counting errors unless heading for a lower limit
	  // switch or the stow position (which may be the same)
	  // but allow one less received count in case there was a
	  // timeout by a motor stopping with reed switch open
	  if ( (rcount + g.get_ptoler() < count) && (g.get_stow() != 1) &&
	      (
		(axis == 0 && g.get_azcmd() != g.get_azlim1()) ||
		(axis == 1 && g.get_elcmd() != g.get_ellim1())
	      ) ) //JDL
	  {
	      d.dtext(16.0, 48.0, gg, Color.red, "lost count");
	      d.dtext(440.0, ylim + 40.0, gg, Color.blue, "COUNT ERROR: -> stow ");
	      d.dtext(8.0, ylim + 60.0, gg, Color.black, "received " +
		  rcount + " counts out of " + count + " counts expected");
	      if (mm == 1)
		  d.dtext(8.0, ylim + 74.0, gg, Color.black, "while increasing azimuth");//JDL
	      if (mm == 0)
		  d.dtext(8.0, ylim + 74.0, gg, Color.black,
		  "while decreasing azimuth");//JDL
	      if (mm == 3)
		  d.dtext(8.0, ylim + 74.0, gg, Color.black, "increasing elevation");//JDL
	      if (mm == 2)
		  d.dtext(8.0, ylim + 74.0, gg, Color.black,
		    "while decreasing elevation");//JDL
	      d.dtext(8.0, ylim + 88.0, gg, Color.black, "motor stalled or limit prematurely reached");
	      if (g.get_fstatus() == 1) {
		  o.stroutfile(g, "* ERROR count = "+count+", rcount = "+rcount);
		  o.stroutfile(g, "* mm = "+mm);
	      }
	      if (lout != null) {
/*
		  try {
		      lout.write(t.getYdhms()+" count error: expected "+count
			+", got "+rcount+"; wrote "+str
			+", read "+rstr);
		      lout.newLine();
		      lout.write("\tget_scan() = "+g.get_scan()+", az = "
			+d.dc(g.get_aznow(), 5, 1)+", el = "+d.dc(g.get_elnow(), 5, 1));
		      lout.newLine();
		      lout.flush();
		  } catch(IOException e) {
		      System.err.println(e);
		  }
*/
		  lout.println(t.getYdhms()+" count error: expected "+count
		    +", got "+rcount+"; wrote "+str
		    +", read "+rstr);
		  lout.println("\tget_scan() = "+g.get_scan()+", az = "
		    +d.dc(g.get_aznow(), 5, 1)+", el = "+d.dc(g.get_elnow(), 5, 1));
		  lout.flush();
	      }
	      // if we get to here, rcount + g.get_ptoler() was < count,
	      // target was not azlim1 or ellim1 (which are also the stow
	      // targets for the cassi mount) or the stow position
	      if (g.get_mainten() == 0) {
		  if (mm == 2 && rstr.charAt(0) == 'T')
		  // el motor timed out decreasing el; could have been while
		  // tracking an astronomical source
		  {
		      g.set_elcount(0);
		      g.set_elnow(g.get_ellim1());
		  }
		  elatstow = azatstow = 0;
		  if (g.get_fstatus() == 1)
		      o.stroutfile(g, "* because of timeout, will return to stow");
		  g.set_clr(1); //JDL
		  g.set_bsw(0); //JDL
		  g.set_scan(0); //JDL
		  g.set_sourn(0); //JDL
		  g.set_track(0); //JDL
		  g.set_azoff(0.0); //JDL
		  g.set_eloff(0.0); //JDL
		  g.set_azcmd(g.get_azstow());
		  g.set_elcmd(g.get_elstow());
		  g.set_stow(1); //JDL
	      }
	      return; // to procs.spectra() -- the only function that calls us
	  }
	  // we are here because there was not a count error to worry
	  // about (see reasons above), but if there was still a timeout;
	  // try to deal with it appropriately

	  // when going to stow, motors time out when the limit switches are
	  // hit, so we don't clutter up the log file reporting such events
	  if (g.get_stow() == 0 && rstr.charAt(0) == 'T' && lout != null) {
/*
	      try {
		  lout.write(t.getYdhms()+" count = "+count+"; wrote "
		      +str+", read "+rstr);
		  lout.newLine();
		  lout.write("\tazcmd = "+d.dc(g.get_azcmd(), 5, 1)
		      +", elcmd = "+d.dc(g.get_elcmd(), 5, 1)
		      +"; az = "+d.dc(g.get_aznow(), 5, 1)
		      +", el = "+d.dc(g.get_elnow(), 5, 1));
		  lout.newLine();
		  lout.write("\tget_scan() = "+g.get_scan()+", get_stow() = "
		    +g.get_stow());
		  lout.newLine();
		  lout.flush();
	      } catch(IOException e) {
		  System.err.println(e);
	      }
*/
	      lout.println(t.getYdhms()+" count = "+count+"; wrote "
		  +str+", read "+rstr);
	      lout.println("\tazcmd = "+d.dc(g.get_azcmd(), 5, 1)
		  +", elcmd = "+d.dc(g.get_elcmd(), 5, 1)
		  +"; az = "+d.dc(g.get_aznow(), 5, 1)
		  +", el = "+d.dc(g.get_elnow(), 5, 1));
	      lout.println("\tget_scan() = "+g.get_scan()+", get_stow() = "
		+g.get_stow());
	      lout.flush();
	  }
	  // assume we have reached stow el
	  if (mm == 2 && rstr.charAt(0) == 'T' && count == 5000) {
	      elatstow = 1;
	      g.set_elcount(0);
	      g.set_elnow(g.get_elstow());
	  }
	  // assume we have reached stow az
	  if (mm == 0 && rstr.charAt(0) == 'T' && count == 5000) {
	      azatstow = 1;
	      g.set_azcount(0);
	      g.set_aznow(g.get_azstow());
	  }
	  // g.set_stow(-1) will be done about 80 lines below
	  // this timeout will have been logged above in error file
	  if (rstr.charAt(0) == 'T' && g.get_stow() == 0) {
	      d.dtext(16.0, 32.0, gg, Color.black, "timeout from antenna");
	  }
	  // this should be the normal situation
	  if (rstr.charAt(0) == 'M') {
	      if (axis == 0) {
		  azatstow = 0;
		  if (mm == 1)
		    g.set_azcount(g.get_azcount() + fcount);
		  else
		    g.set_azcount(g.get_azcount() - fcount);
	      }
	      if (axis == 1) {
		  elatstow = 0;
		  if (mm == 3)
		    g.set_elcount(g.get_elcount() + fcount);
		  else
		    g.set_elcount(g.get_elcount() - fcount);
	      }
	  }
	  if (g.get_azelsim() == 0) {
	      try {
		  Thread.sleep(5);
	      } catch(InterruptedException e) {
		  System.err.println(e);
	      }
	  }
	} // end of if (count > ptoler && track != -1)
	// do the second axis
	axis++;
	if (axis > 1)
	    axis = 0;
    }
    // both axis motions should now be complete
    if (g.get_track() != -1) {
	if (g.get_slew() == 1)
	    g.set_track(0);
	else
	    g.set_track(1);
    }
    g.set_aznow(g.get_azlim1() - g.get_azcor() + g.get_azcount() * 0.5 / azscale);
    if (g.get_aznow() > 360.0)
	g.set_aznow(g.get_aznow() - 360.0);
    if (g.get_pushrod() == 0)
	ellnow = g.get_elcount() * 0.5 / elscale;
    else {
	ellnow = -g.get_elcount() * 0.5 / g.get_rod5() + lenzero;
	ellnow = g.get_rod1() * g.get_rod1() + g.get_rod2() * g.get_rod2()
	  - g.get_rod3() * g.get_rod3() - ellnow * ellnow;
	ellnow = ellnow / (2.0 * g.get_rod1() * g.get_rod2());
	ellnow = -Math.acos(ellnow) * 180.0 / Math.PI + g.get_rod4()
	  - g.get_ellim1();
//        System.out.println("ellnow "+ellnow);
    }
    g.set_elnow(g.get_ellim1() - g.get_elcor() + ellnow);
    if (g.get_elnow() > 90.0) {
	if (g.get_aznow() >= 180.0)
	    g.set_aznow(g.get_aznow() - 180.0);
	else
	    g.set_aznow(g.get_aznow() + 180.0);
	g.set_elnow(180.0 - g.get_elnow());
    }
    d.dtext(670.0, 40.0, gg, Color.black,
	"azel  " + d.dc(g.get_aznow(), 5, 1) + " " + d.dc(g.get_elnow(), 5, 1)
	+ " deg");
    x = g.get_xlast(0);
    y = g.get_ylast(0);
    d.lpaint(gg, Color.white, (double) (x - 10), (double) y, (double) (x + 10), (double) y);
    d.lpaint(gg, Color.white, (double) x, (double) (y - 10), (double) x, (double) (y + 10));
    x = (int) (g.get_aznow() * 640.0 / 360.0);
    if (g.get_south() == 0) {
	x -= 320;
	if (x < 0)
	    x += 640;
    }
    y = (int) (ylim - g.get_elnow() * (ylim - ylim0) / 90);
    g.set_xlast(x, 0);
    g.set_ylast(y, 0);
    d.lpaint(gg, Color.red, (double) (x - 10), (double) y, (double) (x + 10), (double) y);
    d.lpaint(gg, Color.red, (double) x, (double) (y - 10), (double) x, (double) (y + 10));
//    if (Math.abs(g.get_aznow() - g.get_azlim1()) < 0.1 &&
//      Math.abs(g.get_elnow() - g.get_ellim1()) < 0.1)
// JDL counts/degree is ~4.75 and count tolerance is usually 1, so
// 0.3 seems more appropriate
    if (Math.abs(g.get_aznow() - g.get_azstow()) < 0.3 &&
      Math.abs(g.get_elnow() - g.get_elstow()) < 0.3 )
    {
	d.set_bc(Color.green, 3);
	d.dtext(340.0, ylim + 40.0, gg, Color.black, "Status: at stow");
	g.set_stow(-1);     // at stow
	d.ppclear(gg, 440.0, ylim + 40.0, 100.0); //JDL
    } else {
	d.set_bc(Color.black, 3);
	if (g.get_stow() == -1)
	    g.set_stow(0);
    }
    if (g.get_stow() != 0) {
	g.set_track(0);
    }
    sec = (double) t.getTsec(g, d, gg);
    ra = geom.get_galactic_ra(sec, g.get_aznow(), g.get_elnow(), g, t);
    dec = geom.get_galactic_dec();
    glat = geom.get_galactic_glat();
    glon = geom.get_galactic_glon();
    vel = geom.get_galactic_vel();
    if (Math.abs(g.get_fcenter() - 1665.4) < 1.0)
	g.set_restfreq(1665.4);
    else
	g.set_restfreq(1420.406);
    if (Math.abs(g.get_fcenter() - 1612.231) < 1.0)
	g.set_restfreq(1612.231);
    g.set_vlsr(vel);
    g.set_glat(glat);
    g.set_glon(glon);
    if (Math.abs(g.get_fcenter() - 1667.359) < 1.0)
	g.set_restfreq(1667.359);
    d.dtext(670.0, 120.0, gg, Color.black,
	"radec " + d.dc(ra, 5, 1) + " hrs " + d.dc(dec, 5, 1) + " deg");
    d.dtext(670.0, 104.0, gg, Color.black, "Galactic l =" + d.dc(glon, 4, 0) + " b =" + d.dc(glat, 3, 0));
    d.dtext(665.0, 426.0, gg, Color.black, "VLSR    " + d.dc(vel, 6, 1) + " km/s");
    vel = -vel - (g.get_fcenter() - g.get_restfreq()) * 299790.0 / g.get_restfreq();
    d.dtext(665.0, 441.0, gg, Color.black, "Vcenter " + d.dc(vel, 6, 1) + " km/s");
    d.lpaint(gg, Color.black, 0.0, 85.0, 217.0, 85.0);
    d.stext(16.0, ylim + 22.0, gg, Color.black,
	g.get_statnam() + " lat " + d.dc(g.get_lat() * 180.0 / Math.PI, 4, 1)
	+ " lon " + d.dc(g.get_lon() * (-180.0) / Math.PI, 4, 1));
    d.lpaint(gg, Color.black, 0.0, 0.0, 217.0, 0.0);
    return; // to the main loop in srt.java
  } // end of azel()

    // communicates with the digital radiometer
    double radiodg(double freq, global g, disp d, Graphics gg, outfile o)
    {
        double power, avpower, tsig, a;
        int k, n, j, mode, i;
	int counts[] = new int[64];
        byte b8, b9, b10, b11;
	String req;
	String rets = "";
	String[] part;
        power = tsig = 0.0;

	if (g.get_elog()) { //JDL
	  if ((lout = g.get_elout()) == null) {
	    lout = g.open_elog();
	  }
	} else {
	  lout = null;
	}
        j = (int) (freq * (1.0 / 0.04) + 0.5); /* bits for reference divider of syn */
        mode = g.get_digital() - 1;
        if (g.get_digital() == 4 || g.get_digital() == 5)
            mode = 0;
	b8 = (byte) mode;
        b9 = (byte) (j & 0x3f);
        b10 = (byte) ((j >> 6) & 0xff);
        b11 = (byte) ((j >> 14) & 0xff);
        g.set_freqa(((b11 * 256.0 + (b10 & 0xff)) * 64.0 + (b9 & 0xff)) * 0.04 - 0.8);
	// no space at start, but a space needed after each number
	req = new String("radio " + j + " " + mode + "  ");
	n = 0;
        if (g.get_radiosim() == 0) {
            try {
		outWriter.println(req);
                try {
		    spSocket.setSoTimeout(25000); // 25s should be plenty
                }
		catch(SocketException e) {
		    System.err.println("radiodg: can't set timeout "+e);
                }
		try {
//                    rets = new String(inReader.readLine());
		    rets = inReader.readLine();
		    if (rets.startsWith("BEEP")) {
			g.set_beep(true);
//                        rets = new String(inReader.readLine());
			rets = inReader.readLine();
		    }
		    outWriter.println("GotIt");
		} catch(NullPointerException np) {
		    rets = "";
		} catch(SocketTimeoutException e) { // extends InterruptedIOException
		    System.err.println("radiodg: socket read() timeout");
		}
		part = rets.split("\\s", 0);
		if (rets.length() == 0) {
		    n = 0;
		} else {
		    n = part.length;
		}
		if (n != 66) {
		    System.err.println("radio sent back "+n+" integers");
/*
		    if (lout != null) {
		      try {
			lout.write("Radio error (only "+n+" integers).");
			lout.newLine();
			lout.flush();
		      } catch(IOException e) {
			System.err.println(e);
		      }
		    }
*/
		    lout.println("Radio error (only "+n+" integers).");
		    lout.flush();
		}
		try {
		    if (n > 0)
			n = Integer.parseInt(part[0]);
		} catch(NumberFormatException fe) {
		    n = 0;
		}
		if (n == 128) {
		    for (i = 0; i < 64; i++) {
			try {
			    counts[i] = Integer.parseInt(part[i+1]);
			} catch(NumberFormatException fe) {
			    System.err.println("Error parsing radio data "+i);
			    counts[i] = 0;
			}
		    }
		}
            }
            catch(IOException e) {
                t.getTsec(g, d, gg);
                d.ppclear(gg, 16.0, 48.0, 180.0);
		d.dtext(16.0, 48.0, gg, Color.red, "waiting on recvr");
		System.err.println(e);
            }
        }

        if (n != 128 && g.get_radiosim() == 0) {
	    d.dtext(16.0, 16.0, gg, Color.red, "error comm with radio, n =  " + n);
            g.set_comerad(g.get_comerad() + 1);
            if (g.get_fstatus() == 1)
                o.stroutfile(g, "* ERROR communicating with radio");
	    if (lout != null) {
/*
		try {
		  lout.write(t.getYdhms()+" Radio error (n = "+n+").");
		  lout.newLine();
		  lout.flush();
		} catch(IOException e) {
		  System.err.println(e);
		}
*/
		lout.println(t.getYdhms()+" Radio error (n = "+n+").");
		lout.flush();
	    }
	    return(-1.0);
        } else {
            a = avpower = 0;
            for (i = 0; i < 64; i++) {
                if (g.get_radiosim() != 0) {
                    power = 200.0 + g.get_tspill();
                    if (g.get_sourn() > 0 && i == 0) {
                        if (g.get_sounam(g.get_sourn()).lastIndexOf("Moon") > -1)
                            tsig = 1.0;
                        if (g.get_sounam(g.get_sourn()).lastIndexOf("Cass") > -1)
                            tsig = 2.6;
                        if (g.get_sounam(g.get_sourn()).lastIndexOf("Sun") > -1)
                            tsig = 1000.0;
                    }
                    if (g.get_scan() != 0 || g.get_track() != 0) {
                        a = g.get_azoff() * Math.cos(g.get_elcmd() * Math.PI / 180.0);
                        a = (g.get_eloff() * g.get_eloff() + a * a) * 4.0 * 0.7;
                        a = a / (g.get_beamw() * g.get_beamw());
                        power += tsig * Math.pow(2.718, -a);
                    }
                    try {
                        Thread.sleep(5);
		    } catch(InterruptedException e) {
                        System.out.println(e);
                    }
                    if (g.get_calon() == 1)
                        power = power + g.get_tload() - g.get_tspill();
                    if (g.get_calon() == 2)
                        power = power + g.get_noisecal();
                    power += power * gauss() / Math.sqrt(g.get_freqsep() * 1e06 * g.get_intg());
                    d.dtext(16.0, 80.0, gg, Color.black, "generating random data");
                    power = power * g.get_graycorr(i);
                }
                if (g.get_radiosim() == 0) {
		    power = (double)counts[i];
                }


		// offset from center frequency
                if (g.get_digital() < 5)
                    a = (i - 32) * g.get_freqsep() * 0.4;
                else
                    a = 0;

		// The controller encodes 64 16-bit integers in a 128-byte
		// array recv[]; SRTsocket combines them to make the
		// 64 integers and returns them in an ascii string.
		// See the comments in SRTsocket.java.

		// This if() excludes i < 8 and i > 56; thus recv[j]
		// where 48 < j < 113 are not corrected; however the
		// graycorr correction is applied to all the central
		// 44 frequencies used in the average power calculation.
                if (g.get_graycorr(i) > 0.8)
                    power = power / (g.get_graycorr(i) * (1.0 + a * a * g.get_curvcorr()));

                a = g.get_calcons() * power;

		// double specd[64] stores the counts for each of the 64
		// frequency channels, and 2nd arg to g.set_specd() is the
		// index into specd[]
		// The components of specd[] are related to recv[] by:
		//    specd[0]  from recv[64] & recv[65] (counts[0])
		//    specd[1]  from recv[62] & recv[63] (counts[63])
		//    specd[2]  from recv[60] & recv[61] (counts[62])
		//    ...
		//    specd[11] from recv[42] & recv[43] (counts[53])
		//    ...
		//    specd[31] from recv[2]  & recv[3]  (counts[33])
		//    specd[32] from recv[0]  & recv[1]  (counts[32])
		//    specd[33] from recv[126] & recv[127] (counts[31])
		//    ...
		//    specd[54] from recv[84] & recv[85] (counts[10])
		//    ...
		//    specd[63] from recv[66] & recv[67] (counts[01])

                if (i > 0)
		    g.set_specd(a, 64 - i); // reverse lower sideband

                else
                    g.set_specd(a, 0);
		if (i >= 10 && i < 54) // specd[[54] to specd[11]
                    avpower += power;
            }
            avpower = avpower / 44.0;
            a = g.get_calcons() * avpower;
            d.dtext(8.0, ylim + 40.0, gg, Color.black, "digital Recvr freq: " +
                    d.dc(g.get_freqa(), 7, 2) + " pwr: "
                    + d.dc(avpower, 4, 0) + " counts" + " temp: " + d.dc(a, 4, 0) + "K");
	    return(a);
        }
    }   // end of radiodg()

    double gauss() {
        double v1, v2, r, fac, amp, vv1;
        r = v1 = 0.0;
        while (r > 1.0 || r == 0.0) {
            v1 = 2.0 * Math.random() - 1.0;
            v2 = 2.0 * Math.random() - 1.0;
            r = v1 * v1 + v2 * v2;
        }
        fac = Math.sqrt(-2.0 * Math.log(r) / r);
        vv1 = v1 * fac;
        amp = vv1;
	return(amp);
    }

    // move the calibration vane or control the noise diode
    // mode==0 calout (B8==4),  mode==1 calin (B8==5)
    // mode==2 ncaloff (B8==6), mode==3 ncalon (B8==7)
    void cal(int mode, global g, disp d, Graphics gg, outfile o)
    {

        int j, k, kk, n, i;
	String str, str2;
	String rstr = null;

	if (g.get_elog()) { //JDL
	  if ((lout = g.get_elout()) == null) {
	    lout = g.open_elog();
	  }
	} else {
	  lout = null;
	}

	d.ppclear(gg, 16.0, 32.0, 100.0);
        if (g.get_azelsim() == 0)
	    d.dtext(16.0, 32.0, gg, Color.black, "calib. status:");
        else
	    d.dtext(16.0, 32.0, gg, Color.black, "calib. simulated");
	// no space at start, but a space needed after each number
	str = "move " + (mode + 4) + " 0  ";

	d.dtext(16.0, 64.0, gg, Color.black, "motor: " + str + "     ");
        d.ppclear(gg, 16.0, 80.0, 180.0);

        if (g.get_azelsim() == 0) {
            try {
		outWriter.println(str);
                try {
		    spSocket.setSoTimeout(20000);
		} catch(SocketException e) {
		    System.err.println("cal: can't set timeout "+e);
                }
		if ((mode & 1) == 1)
		    d.dtext(16.0, 48.0, gg, Color.black, "waiting on calin/on  ");
		else
		    d.dtext(16.0, 48.0, gg, Color.black, "waiting on calout/off ");
		try {
		    Thread.sleep(2000);
		} catch(InterruptedException e) {
		    ;
		}
		try {
		    rstr = new String(inReader.readLine());
		    while (rstr.startsWith("BEEP")) {
			g.set_beep(true);
			rstr = new String(inReader.readLine());
		    }
		    outWriter.println("GotIt");
		} catch(SocketTimeoutException e) { // extends InterruptedIOException
		    System.out.println("cal: read() timeout");
		}
		if (rstr == null || rstr.length() < 4)
		    n = -1;
		else
		    n = rstr.length();
                d.ppclear(gg, 16.0, 48.0, 180.0);
	    } catch(IOException e) {
		n = -1;
		System.err.println(e);
            }
	} else {  // simulation
	    rstr = new String("C 0 " +(mode+4)+" 0 ");
	    n = rstr.length();
        }
	if (n == -1 && g.get_azelsim() == 0) {
	    d.dtext(16.0, 16.0, gg, Color.red, "comerr: n=" + n);
            g.set_comerr(g.get_comerr() + 1);
            if (g.get_fstatus() == 1)
                o.stroutfile(g, "* ERROR comerr on cal");
	    if (lout != null) {
/*
		try {
		  lout.write(t.getYdhms()+" Cal com error, "+str);
		  lout.newLine();
		  lout.flush();
		} catch(IOException e) {
		  System.err.println(e);
		}
*/
		lout.println(t.getYdhms()+" Cal com error, "+str);
		lout.flush();
	    }
            return;
        }
	d.dtext(16.0, 80.0, gg, Color.black, "reply: " + rstr);
        return;
    }

    // called on exit
    void close(global g) {
	if (g.get_azelsim() == 0 || g.get_radiosim() == 0)
	    try {       // closes log file & the socket stuff
		if (g.get_elout() != null) {
		    g.close_elog();
		    lout = null;
		}
		outWriter.println("bye. ");
		try {
		    Thread.sleep(1000);
		} catch(InterruptedException e) {
		    ;
		}
		outWriter.close();
		inReader.close();
		spSocket.close();
	    } catch(IOException e) {
		System.err.println("Closing sockport: "+e);
	    }
    }
}
